function love.load()	
	require("piece_square")
	--Class and board initialization.
	game = {}
	game.turn = 0
	
	board = {}

	--[[
		BOARD
	]]

	board.file = {"a", "b", "c", "d", "e", "f", "g", "h"}

	function board.new()
		for file = 1, 8 do
			board[file] = {}
			local is_white = file % 2 == 0 and true or false

			for rank = 8, 1, -1 do
				is_white = not is_white
				board[file][rank] = square(is_white and "white" or "black", file, rank)
			end
		end
	end

	board.new()

	for i = 1, 8 do
		pawn("black", board[i][7])
		pawn("white", board[i][2])
	end

	rook("black", board[1][8])
	knight("black", board[2][8])
	bishop("black", board[3][8])
	queen("black", board[4][8])
	king("black", board[5][8])
	bishop("black", board[6][8])
	knight("black", board[7][8])
	rook("black", board[8][8])
	
	rook("white", board[1][1])
	knight("white", board[2][1])
	bishop("white", board[3][1])
	queen("white", board[4][1])
	king("white", board[5][1])
	bishop("white", board[6][1])
	knight("white", board[7][1])
	rook("white", board[8][1])
	
	--[[
	
		LOVE
	
	]]
	
	love.window.setTitle("Silver Chess")
	love.window.setIcon(love.image.newImageData("logo.png"))

	love.window.setMode(100, 100, {resizable = true, minwidth = 512, minheight = 512})
	scr_w, scr_h = 512, 512
	
	sound = {}
	sound.move = love.audio.newSource("move.wav", "static")
	sound.victory = love.audio.newSource("victory.mp3", "stream")
	sound.stalemate = love.audio.newSource("stalemate.wav", "static")
	
	draw_chessboard()
	draw_square_names()

	images = {}
	images.pieces = love.graphics.newImage("chess_set.png")
	
	images.white = {}
	images.black = {}
	
	images.black.king = love.graphics.newQuad(0, 0, 64, 64, images.pieces:getDimensions())
	images.black.queen = love.graphics.newQuad(64, 0, 64, 64, images.pieces:getDimensions())
	images.black.rook = love.graphics.newQuad(128, 0, 64, 64, images.pieces:getDimensions())
	images.black.bishop = love.graphics.newQuad(192, 0, 64, 64, images.pieces:getDimensions())
	images.black.knight = love.graphics.newQuad(256, 0, 64, 64, images.pieces:getDimensions())
	images.black.pawn = love.graphics.newQuad(320, 0, 64, 64, images.pieces:getDimensions())

	images.white.king = love.graphics.newQuad(0, 64, 64, 64, images.pieces:getDimensions())
	images.white.queen = love.graphics.newQuad(64, 64, 64, 64, images.pieces:getDimensions())
	images.white.rook = love.graphics.newQuad(128, 64, 64, 64, images.pieces:getDimensions())
	images.white.bishop = love.graphics.newQuad(192, 64, 64, 64, images.pieces:getDimensions())
	images.white.knight = love.graphics.newQuad(256, 64, 64, 64, images.pieces:getDimensions())
	images.white.pawn = love.graphics.newQuad(320, 64, 64, 64, images.pieces:getDimensions())
end

function love.update(dt)
	--require("lovebird").update()
	
	--love.window.setTitle(game.turn % 2 == 0 and "White's Turn!" or "Black's Turn!")
	scale = size / 64
	mouse_x, mouse_y = love.mouse.getPosition()
	
	hovering_square = {x = -1, y = -1}
	
	if not love.mouse.isDown(1) then
		dragging_square = nil
	else
		dragging_sqaure = dragging_sqaure
	end
end

function love.mousereleased(x, y, button)
	if button == 1 then
		if promotion then
			local promotion_file = promotion.s:GetFile()
			local enemy_color = promotion.c == "white" and "black" or "white"
			
			for rank, piece in pairs(promotion.b) do
				if (8 - hovering_square.y) == rank and (1 + hovering_square.x) == promotion_file then
					piece(promotion.c, promotion.s)
					promotion.o:SetPiece(nil)
					promotion = nil
					game.turn = game.turn + 1
					love.audio.play(sound.move)
					
					if is_checkmated(enemy_color) then
						love.window.setTitle(promotion.c .. " wins by checkmate.")
						love.audio.play(sound.victory)
					end
					
					if is_stalemated(enemy_color) then
						love.window.setTitle(enemy_color .. " has been stalemated.")
						love.audio.play(sound.stalemate)
					end
				end
			end
		end
	end

	if button == 2 then
		promotion = nil
	end

	if not dragging_square then
		return
	end
	
	if not hovering_square then
		return
	end
	
	if button == 1 and not promotion then --prevents us from moving a different piece while trying to promote and then getting a free promotion
		board[dragging_square.x + 1][8 - dragging_square.y]:Move(board[hovering_square.x + 1][8 - hovering_square.y])
	end
end

function is_checkmated(color)
	local enemy_color = color == "white" and "black" or "white"
	local king_square = get_king_square(color)

	
	if not is_square_attacked(enemy_color, king_square) then
		return false
	end

	for x = 1, 8 do
		for y = 1, 8 do
			local square = board[x][y]
			local piece = square:GetPiece()
			
			if piece and piece:GetColor() == color then
				for file = 1, 8 do
					for rank = 1, 8 do
						local new_square = board[file][rank]
						local new_piece = new_square:GetPiece()
						
						if piece:IsMoveLegal(new_square) then
							return false --we aren't checkmated because there is a legal move.
						end
					end
				end
			end
		end
	end
	
	return true
end

function is_stalemated(color)
	local enemy_color = color == "white" and "black" or "white"
	local king_square = get_king_square(color)

	
	if is_square_attacked(enemy_color, king_square) then --Different from checkmate because you cannot be in check to be stalemated
		return false
	end
	
	for x = 1, 8 do
		for y = 1, 8 do
			local square = board[x][y]
			local piece = square:GetPiece()
			
			if piece and piece:GetColor() == color then
				for file = 1, 8 do
					for rank = 1, 8 do
						local new_square = board[file][rank]
						local new_piece = new_square:GetPiece()
						
						if piece:IsMoveLegal(new_square) then
							return false --we aren't checkmated because there is a legal move.
						end
					end
				end
			end
		end
	end
	
	return true
end


function draw_promotion()
	promotion_squares = love.graphics.newCanvas(scr_w, scr_h)
	love.graphics.setCanvas(promotion_squares)
	size = scr_w > scr_h and scr_h / 8 or scr_w / 8
	
	local promotion_square = promotion.s
	local promotion_color = promotion.c
	local promotion_b = promotion.b
	local promotion_file = promotion_square:GetFile()
	
	for y = 0, 7 do
		for x = 0, 7 do
			local square = board[x + 1][8 - y]
			
			if x + 1 == promotion_file then --If we are on the same file...
				--promotion = {o = cur_square, c = self:GetColor(), s = square, b = {1 = queen, 2 = rook, 3 = knight, 4 = bishop}}
				for rank, piece in pairs(promotion_b) do
					if rank == (8 - y) then
						love.graphics.setColor(255, 255, 255)
						love.graphics.rectangle("fill", x * size, y * size, size, size)
						love.graphics.draw(images.pieces, images[promotion_color][piece.mt.name], x * size, y * size, 0, scale, scale) --maybe problems here
					end
				end
			end
		end
	end
	
	love.graphics.setCanvas()
end

function love.resize(w, h)
	local smaller = w > h and h or w
	scr_w, scr_h = smaller, smaller
	
	love.window.setMode(smaller, smaller, {resizable = true, minwidth = 512, minheight = 512})
	draw_chessboard()
	draw_square_names()
end

function draw_chessboard()
	background_squares = love.graphics.newCanvas(scr_w, scr_h)
	love.graphics.setCanvas(background_squares)
	size = scr_w > scr_h and scr_h / 8 or scr_w / 8
	
		for y = 0, 7 do
			local is_white = true
			
			if y % 2 ~= 0 then
				is_white = not is_white
			end
			
			for x = 0, 7 do
				if is_white then
					love.graphics.setColor(255, 255, 255)
				else
					love.graphics.setColor(0, 0, 0)
				end
				
				local x_pos = x * size
				local y_pos = y * size
				
				love.graphics.rectangle("fill", x_pos , y_pos, size, size) --Draw appropriately colored square.
				
				is_white = not is_white
			end
		end
	love.graphics.setCanvas()
end

function is_square_attacked(attacking_color, attacked_square) --Not good for drawing, king castling stuff messes that up...	
	for x = 1, 8 do
		for y = 1, 8 do
			local square = board[x][y]
			local piece = square:GetPiece()
			
			if piece and piece:GetColor() == attacking_color then
				if piece:CanMove(attacked_square) then
					return true
				end
			end
		end
	end
end

function get_king_square(color) --Assumes there is only 1 king of any given color on the board.
	for x = 1, 8 do	
		for y = 1, 8 do
			local square = board[x][y]
			local piece = square:GetPiece()
			
			if piece and piece:GetName() == "king" and piece:GetColor() == color  then --Wow...
				return square
			end
		end
	end
end

function draw_moves()
	move_squares = love.graphics.newCanvas(scr_w, scr_h)
	love.graphics.setCanvas(move_squares)
		local d_p = dragging_square.p

		for y = 0, 7 do
			for x = 0, 7 do
				if d_p:IsMoveLegal(board[x + 1][8 - y]) then
					local square = board[x + 1][8 - y]
					local piece = square:GetPiece()

					if piece then
						love.graphics.setColor(255, 0, 0, 100)
					else
						love.graphics.setColor(0, 255, 0, 100)
					end
					
					love.graphics.rectangle("fill", x * size, y * size, size, size)
				end
			end
		end
		
	love.graphics.setCanvas()
end

function draw_square_names()
	square_names = love.graphics.newCanvas(scr_w, scr_h)
	love.graphics.setCanvas(square_names)
	
	local scale = scr_w / 8 / 64
	
		for y = 0, 7 do
			local is_white = true
			if y % 2 ~= 0 then
				is_white = not is_white
			end
			
			for x = 0, 7 do
				if is_white then
					love.graphics.setColor(0, 0, 0)
				else
					love.graphics.setColor(255, 255, 255)
				end
				
				local x_pos = x * size
				local y_pos = y * size
				
				if x == 0 then
					love.graphics.print(8 - y, x_pos + 1 , y_pos + 1, 0, scale, scale)
				end
				
				if y == 7 then
					local letter = love.graphics.newText(love.graphics.getFont(), board.file[x + 1])
					love.graphics.draw(letter, x_pos + size - letter:getWidth() * scale - 1, y_pos + size - letter:getHeight() * scale - 1, 0, scale, scale)
				end
				
				is_white = not is_white
			end
		end
	love.graphics.setCanvas()
end

function love.draw()
	love.graphics.setColor(255, 255, 255)
	love.graphics.draw(background_squares, 0, 0)
	love.graphics.draw(square_names, 0, 0)

	for y = 0, 7 do
		for x = 0, 7 do
			local x_pos = x * size
			local y_pos = y * size
						
			local sq = board[x + 1][8 - y]
			local piece = sq:GetPiece()
			
			--Makes sure our mouse is within the borders of our square.
			if mouse_x >= x_pos and mouse_y >= y_pos and mouse_x <= x_pos + size and mouse_y <= y_pos + size then 
				hovering_square = {x = x, y = y}
				
				if love.mouse.isDown(1) and piece and not dragging_square then
					dragging_square = {x = x, y = y, p = piece} --If we are holding left click on a square with a piece and we aren't dragging a piece, now we are.
					draw_moves()
				end
			end
					
			if piece then
				if not dragging_square or dragging_square.p ~= piece then --Render a piece only if we aren't dragging it.
					love.graphics.setColor(255, 255, 255)
					love.graphics.draw(images.pieces, images[piece:GetColor()][piece:GetName()], x_pos, y_pos, 0, scale, scale)
				end
			end
		end
	end
	
	if promotion then
		love.graphics.setColor(255, 255, 255)
		draw_promotion()
		love.graphics.draw(promotion_squares, 0, 0)
	elseif dragging_square then
		local d_p = dragging_square.p
		local color = d_p:GetColor()
		
		if color == "black" and game.turn % 2 ~= 0 or color == "white" and game.turn % 2 == 0 then
			love.graphics.setBlendMode("alpha", "premultiplied")
			love.graphics.setColor(255, 255, 255)
			love.graphics.draw(move_squares, 0, 0)
		end
		
		love.graphics.setBlendMode("alpha")
		
		if hovering_square then --Draws a highlighted square over the square that we are trying to drag a piece to.
			love.graphics.setColor(255, 255, 255, 100)
			love.graphics.rectangle("fill", hovering_square.x * size, hovering_square.y * size, size, size)
		end
		
		love.graphics.setColor(255, 255, 255)
		love.graphics.draw(images.pieces, images[d_p:GetColor()][d_p:GetName()], mouse_x - size / 2, mouse_y - size / 2, 0, scale, scale)
	end
end

--[[
	known "bugs": En Passante captures show up as green instead of red because there is no piece at the capture square.
		
	Make it excecutable
	
	Sound
	Networking
]]


