square = {}

square.mt = {}
square.mt.__index = square.mt

square.mt.piece = nil
square.mt.color = "white"

setmetatable(square, square.mt)

function square.mt:__call(color, file, rank)
	local s = setmetatable({}, square.mt)
	s.color = color
	s.file = file
	s.rank = rank
	return s
end

function square.mt:GetPiece()
	return self.piece
end

function square.mt:GetColor()
	return self.color
end

function square.mt:SetPiece(piece)
	self.piece = piece
end

function square.mt:SetColor(color)
	self.color = color:lower()
end

function square.mt:CanMove(square)
	local piece = self:GetPiece()
	
	if not piece then 
		return false
	end
	
	if not piece:CanMove(square) then
		return false
	end
	
	return true
end

function square.mt:Move(square)
	local piece = self:GetPiece()
	
	if piece then
		piece:Move(square)
	end
end

function square.mt:GetFile()
	return self.file
end

function square.mt:GetRank()
	return self.rank
end

function square.mt:GetUp(count)
	local file = self:GetFile()
	local rank = self:GetRank()
	local new_rank = rank + count
	
	if (new_rank < 9) and (new_rank > 0) then
		return board[file][new_rank]
	end
end

function square.mt:GetRight(count)
	local file = self:GetFile()
	local rank = self:GetRank()
	local new_file = file + count
	
	if (new_file < 9) and (new_file > 0) then
		return board[new_file][rank]
	end
end
--[[
	PIECE
]]

piece = {}

piece.mt = {}
piece.mt.__index = piece.mt

piece.mt.name = nil
piece.mt.color = "white"
piece.mt.square = nil
piece.mt.move_count = 0
piece.mt.last_move = {x = nil, y = nil}

function piece.mt:GetName()
	return self.name
end

function piece.mt:GetColor()
	return self.color
end

function piece.mt:GetSquare()
	return self.square
end

function piece.mt:GetMoveCount()
	return self.move_count
end

function piece.mt:SetName(name)
	self.name = name:lower()
end

function piece.mt:SetColor(color)
	self.color = color:lower()
end

function piece.mt:SetSquare(square)
	self.square = square
end

function piece.mt:SetMoveCount(move_count)
	self.move_count = move_count
end

function piece.mt:GetLastMove()
	return self.last_move
end

function piece.mt:CanMove(square)

end

function piece.mt:IsMoveLegal(square)
	local cur_square = self:GetSquare()
	
	if self:GetSquare() == square then
		return --May end up being redundant. CanMove function should check for this
	end
	
	if self:GetColor() == "white" and game.turn % 2 ~= 0 then --Let's us move or not move based on whose turn it is.
		return
	elseif self:GetColor() == "black" and game.turn % 2 == 0 then
		return
	end
	
	
	if self:CanMove(square) then
		local piece = square:GetPiece()
		
		self:SetSquare(square)
		square:SetPiece(self)
		cur_square:SetPiece(nil)
		
		if is_square_attacked(self:GetColor() == "white" and "black" or "white", get_king_square(self:GetColor())) then
			self:SetSquare(cur_square)  --In case we return false, we return these values back to what they were.
			square:SetPiece(piece)
			cur_square:SetPiece(self)
			
			return false
		end

		self:SetSquare(cur_square)
		square:SetPiece(piece)
		cur_square:SetPiece(self)
		
		if self:GetName() == "pawn" then 
			local last_piece = last_piece_to_move
			local rank_difference = square:GetRank() - cur_square:GetRank()
			local file_difference = square:GetFile() - cur_square:GetFile()
			local is_positive = file_difference > 0
			
			if square:GetRank() == 8 then
				--promotion = {o = cur_square, c = self:GetColor(), s = square, b = {[8] = queen, [7] = rook, [6] = knight, [5] = bishop}}
				return {promotion = true, p = {o = cur_square, c = self:GetColor(), s = square, b = {[8] = queen, [7] = rook, [6] = knight, [5] = bishop}}}
			elseif square:GetRank() == 1 then
				--promotion = {o = cur_square, c = self:GetColor(), s = square, b = {[1] = queen, [2] = rook, [3] = knight, [4] = bishop}}
				return {promotion = true, p = {o = cur_square, c = self:GetColor(), s = square, b = {[1] = queen, [2] = rook, [3] = knight, [4] = bishop}}}
			end
			
			if last_piece then
				if last_piece:GetColor() ~= self:GetColor() and last_piece:GetName() == "pawn" and last_piece:GetMoveCount() == 1 then
					local adjacent = self:GetSquare():GetRight(is_positive and 1 or -1)
					
					if adjacent and adjacent:GetPiece() == last_piece then
						local pawn = adjacent:GetPiece()
						
						if pawn then
							local is_positive = rank_difference > 0 --Trying to fix a rare bug where a pawn is captured wrongfully due to en passente.
							
							if adjacent:GetUp(is_positive and 1 or -1) == square then --If the square behind the pawn we are capturing is where our pawn will move...
								return {en_passante = true, adjacent = adjacent}
							end
						end
					end		
				end
			end
		end
		
		if self:GetName() == "king" then
			local rank_difference = square:GetRank() - cur_square:GetRank()
			local file_difference = square:GetFile() - cur_square:GetFile()
			local positive_file = file_difference > 0
			
			if rank_difference == 0 and (file_difference == 2 or file_difference == -2) then --If we are castling...
				local rook_square = square:GetRight(positive_file and 1 or -2) --We shouldn't have to validate that the rook is there, CanMove should already do that.
				local rook = rook_square:GetPiece()
				local new_rook_square = square:GetRight(positive_file and -1 or 1)
				
				if is_square_attacked(self:GetColor() == "white" and "black" or "white", cur_square) then
					return false --stop fucking castling throug check, bitch!
				end
				
				for i = 0, ((file_difference ^ 2) ^ 0.5) do
					local adjacent_square = get_king_square(self:GetColor()):GetRight(positive_file and i or i * -1)
					local piece = adjacent_square:GetPiece()
					
					if is_square_attacked(self:GetColor() == "white" and "black" or "white", adjacent_square) then
						return false
					end
					
					if i ~= 0 and piece then
						return false
					end
					
					if not positive_file then
						if adjacent_square:GetRight(-1):GetPiece() then
							return false --If there is a stupid knight int he way, we can't castle...
						end
					end
				end
			
				return {castling = true, rook = rook, new_rook_square = new_rook_square, old_rook_square = rook_square}
			end
		end
		
		return true --Move the piece now
	end
end

function piece.mt:Move(square)
	local move_table = self:IsMoveLegal(square)
	local color = self:GetColor()
	local enemy_color = color == "white" and "black" or "white"
	
	if move_table then
		if type(move_table) == "table" then
			if move_table.en_passante then
				local adjacent = move_table.adjacent
				adjacent:SetPiece(nil)
			elseif move_table.castling then
				local rook = move_table.rook
				local new_rook_square = move_table.new_rook_square
				local old_rook_square = move_table.old_rook_square
				
				rook:SetSquare(new_rook_square)
				new_rook_square:SetPiece(rook)
				rook:SetMoveCount(rook:GetMoveCount() + 1)
				
				old_rook_square:SetPiece(nil)
			elseif move_table.promotion then
				promotion = move_table.p
				return
			end
		end
	
		self:GetSquare():SetPiece(nil) --Removes the piece from it's previous square.
	
		self:SetSquare(square)
		square:SetPiece(self)
		self:SetMoveCount(self:GetMoveCount() + 1)
		
		last_piece_to_move = self
		game.turn = game.turn + 1
		love.audio.play(sound.move)
				
		if is_checkmated(enemy_color) then
			love.window.setTitle(color .. " wins by checmate.")
			love.audio.play(sound.victory)
		end
		
		if is_stalemated(enemy_color) then
			love.window.setTitle(enemy_color .. " has been stalemated.")
			love.audio.play(sound.stalemate)
		end
	end
end

--[[
	PAWN
]]

pawn = {}

pawn.mt = setmetatable({}, piece.mt)
pawn.mt.__index = pawn.mt

pawn.mt.name = "pawn"

setmetatable(pawn, pawn.mt)

function pawn.mt:__call(color, square)
	local p = setmetatable({}, pawn.mt)
	p:SetColor(color)
	p:SetSquare(square)
	square:SetPiece(p)
end

function pawn.mt:Promote(square, piece) --piece is our namespace and will be called as a function
	local p = piece(self:GetColor(), square)
	
	self:GetSquare():SetPiece(nil) --Removes the piece from it's previous square.

	promotion = false --We are no longer going to draw our promotion options.
	last_piece_to_move = p
	game.turn = game.turn + 1
end

function pawn.mt:CanMove(square) 
	local color = self:GetColor()
	local cur_square = self:GetSquare()
	local move_count = self:GetMoveCount()
	local square_piece = square:GetPiece()
	
	if square_piece then
		if square_piece:GetColor() == color then
			return false
		end
	end
	
	local file_difference = square:GetFile() - cur_square:GetFile()
	local rank_difference = square:GetRank() - cur_square:GetRank()
	
	if rank_difference > 2 or rank_difference < -2 then
		return false --Maximum vertical movement is two squares, only on the first move though.
	end
	
	if file_difference > 1 or file_difference < -1 then
		return false --Maximum horizontal movement is limited to one square.
	end
	
	if move_count > 0 and rank_difference > 1 or move_count > 0 and rank_difference < - 1 then
		return false --Prevents us from moving more than two spaces on second and later moves.
	end
	
	if (rank_difference > 1 or rank_difference < -1) and file_difference ~= 0 then
		return false --Prevents us from moving two forward and one to the left or right.
	end
	
	if (rank_difference > -1 and color == "black") or (rank_difference < 1 and color == "white") then
		return false
	end
	
	local positive_file = file_difference > 0
	local positive_rank = rank_difference > 0
	
	for i = 0, (file_difference ^ 2) ^ 0.5 do
		local piece = cur_square:GetRight(positive_file and i or i * -1):GetUp(positive_rank and i or i * -1):GetPiece()
		local adjacent = cur_square:GetRight(positive_file and i or i * -1):GetPiece()
		
		if not piece then --exception en passente
			if adjacent then
				local rank = adjacent:GetSquare():GetRank()
				
				if adjacent:GetName() == "pawn" and adjacent:GetMoveCount() == 1 and last_piece_to_move == adjacent then 
					if rank == 4 or rank == 5 then
						return true
					end
				end
			end
			
			return false
		end
	end
	
	local was_piece = nil
	
	for i = 0, rank_difference, rank_difference < 0 and -1 or 1 do
		local this_square = cur_square:GetUp(i)
		local piece = this_square:GetPiece()
		was_piece = was_piece or piece
		
		if piece and (piece ~= self) and was_piece and file_difference == 0 then
			return false  --Pawn cannot capture or advance forward with a piece in the way, even if its an enemy piece.
		end
	end
	
	return true
end

--[[
	ROOK
]]

rook = {}

rook.mt = setmetatable({}, piece.mt)
rook.mt.__index = rook.mt

rook.mt.name = "rook"

setmetatable(rook, rook.mt)

function rook.mt:__call(color, square)
	local r = setmetatable({}, rook.mt)
	r:SetColor(color)
	r:SetSquare(square)
	square:SetPiece(r)
end

function rook.mt:CanMove(square) --Castling rule. Cannot walk into check. Cannot castle through check.
	local cur_square = self:GetSquare()
	local square_piece = square:GetPiece()
	
	if square_piece then
		if square_piece:GetColor() == self:GetColor() then
			return false
		end
	end
	
	local file_difference = square:GetFile() - cur_square:GetFile() 
	local rank_difference = square:GetRank() - cur_square:GetRank() 

	if rank_difference ~= 0 and file_difference ~= 0 then
		return false
	end
	
	for i = 0, file_difference, file_difference < 0 and -1 or 1  do
		local piece = cur_square:GetRight(i):GetPiece()
		
		if piece and (piece ~= self) then
			if piece:GetSquare() ~= square then
				return false
			end
		end
	end
	
	for i = 0, rank_difference, rank_difference < 0 and -1 or 1 do
		local piece = cur_square:GetUp(i):GetPiece()
		
		if piece and (piece ~= self) then
			if piece:GetSquare() ~= square then
				return false
			end
		end
	end
		
	return true
end

--[[
	KNIGHT
]]

knight = {}

knight.mt = setmetatable({}, piece.mt)
knight.mt.__index = knight.mt

knight.mt.name = "knight"

setmetatable(knight, knight.mt)

function knight.mt:__call(color, square)
	local n	= setmetatable({}, knight.mt)
	n:SetColor(color)
	n:SetSquare(square)
	square:SetPiece(n)
end

function knight.mt:CanMove(square)
	local cur_square = self:GetSquare()
	local square_piece = square:GetPiece()
	
	if square:GetColor() == cur_square:GetColor() then
		return false	--If the color of the square we're trying to get to is the same as our current square's color, we can't move there.
	end
	
	if square_piece then
		if square_piece:GetColor() == self:GetColor() then
			return false	--If there is a piece at the square we are trying to move to and if it's the same color as our piece's, we can't move there.
		end
	end
	
	local file_difference = square:GetFile() - cur_square:GetFile() 
	local rank_difference = square:GetRank() - cur_square:GetRank()
	
	if (file_difference > 2) or (file_difference < -2) or (file_difference == 0) then
		return false
	end
	
	if (rank_difference > 2) or (rank_difference < -2) or (rank_difference == 0) then
		return false
	end
	
	return true 
end

--[[
	BISHOP
]]

bishop = {}

bishop.mt = setmetatable({}, piece.mt)
bishop.mt.__index = bishop.mt

bishop.mt.name = "bishop"

setmetatable(bishop, bishop.mt)

function bishop.mt:__call(color, square)
	local b = setmetatable({}, bishop.mt)
	b:SetColor(color)
	b:SetSquare(square)
	square:SetPiece(b)
end

function bishop.mt:CanMove(square)
	local cur_square = self:GetSquare()
	local square_piece = square:GetPiece()
	
	if cur_square:GetColor() ~= square:GetColor() then
		return false
	end
	
	if square_piece then
		if square_piece:GetColor() == self:GetColor() then
			return false
		end
	end
	
	local file_difference = square:GetFile() - cur_square:GetFile()
	local rank_difference = square:GetRank() - cur_square:GetRank()
	
	if file_difference ^ 2 ~= rank_difference ^ 2 then
		return false
	end
	
	local positive_file = file_difference > 0
	local positive_rank = rank_difference > 0
	
	for i = 0, (file_difference ^ 2) ^ 0.5 do
		local piece = cur_square:GetRight(positive_file and i or i * -1):GetUp(positive_rank and i or i * -1):GetPiece()
		
		if piece and (piece ~= self) then
			if piece:GetSquare() ~= square then --You can't get to the square you're trying to get to if their is a piece in the way!
				return false
			end
		end
	end
	
	return true
end

--[[
	QUEEN
]]

queen = {}

queen.mt = setmetatable({}, piece.mt)
queen.mt.__index = queen.mt

queen.mt.name = "queen"

setmetatable(queen, queen.mt)

function queen.mt:__call(color, square)
	local q = setmetatable({}, queen.mt)
	q:SetColor(color)
	q:SetSquare(square)
	square:SetPiece(q)
end

function queen.mt:CanMove(square)
	return rook.mt.CanMove(self, square) or bishop.mt.CanMove(self, square)
end

--[[
	KING
]]

king = {}

king.mt = setmetatable({}, piece.mt)
king.mt.__index = king.mt

king.mt.name = "king"

setmetatable(king, king.mt)

function king.mt:__call(color, square)
	local k = setmetatable({}, king.mt)
	k:SetColor(color)
	k:SetSquare(square)
	square:SetPiece(k)
end

function king.mt:CanMove(square)
	local cur_square = self:GetSquare()
	local square_piece = square:GetPiece()
	
	if square_piece then
		if square_piece:GetColor() == self:GetColor() then
			return false
		end
	end
	
	local file_difference = square:GetFile() - cur_square:GetFile()
	local rank_difference = square:GetRank() - cur_square:GetRank()
	local positive_file = file_difference > 0

	if ((file_difference ^ 2) ^ 0.5 > 1) or ((rank_difference ^ 2) ^ 0.5 > 1) then
		if self:GetMoveCount() == 0 then  --castling
			if rank_difference == 0 and (file_difference == 2 or file_difference == -2) then
				local positive_file = file_difference > 0
				local rook = cur_square:GetRight(file_difference + (positive_file and 1 or -2)):GetPiece()
				
				if rook then
					if rook:GetName() == "rook" and rook:GetMoveCount() == 0 and rook:GetColor() == self:GetColor() then --Rook hasn't moved.
						if cur_square:GetRight(file_difference):GetPiece() == nil and cur_square:GetRight(positive_file and 1 or -1):GetPiece() == nil then
							return true
						end
					end
				end
			end
		end
		
		return false
	end
	
	return true
end